#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: Copyright 2020-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the License); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Compile a neural network model for Arm Ethos-U NPUs."""
import argparse
import glob
import mmap
import os
import sys
import time
from configparser import ConfigParser
from typing import List
from typing import Optional

import flatbuffers

from . import architecture_features
from . import compiler_driver
from . import model_reader
from . import rawdata_writer
from . import scheduler
from . import stats_writer
from . import tflite_writer
from ._version import __version__
from .api import API_VERSION
from .debug_database import DebugDatabase
from .errors import InputFileError
from .errors import VelaError
from .hillclimb_allocation import HillClimbAllocator
from .nn_graph import NetworkType
from .nn_graph import TensorAllocator
from .tensor import MemArea
from .tensor import Tensor
from .tflite.Model import Model
from .tflite_mapping import builtin_operator_map
from .tflite_mapping import builtin_operator_name_map
from .tflite_mapping import optype_to_builtintype
from .tflite_model_semantic import TFLiteSemantic
from .tflite_supported_operators import TFLiteSupportedOperators
from .tosa_model_semantic import TosaSemantic
from .tosa_supported_operators import TosaSupportedOperators
from ethosu import regor

TFLITE_MAGIC = 0x334C4654
TOSA_MAGIC = 0x41534F54


def process(
    input_name, enable_debug_db, arch, model_reader_options, compiler_options, scheduler_options, output_format
):
    if compiler_options.timing:
        start = time.time()

    os.makedirs(compiler_options.output_dir, exist_ok=True)
    output_basename = os.path.join(compiler_options.output_dir, os.path.splitext(os.path.basename(input_name))[0])
    DebugDatabase.show_warnings = enable_debug_db

    nng, network_type = model_reader.read_model(input_name, model_reader_options)

    if not nng:
        raise InputFileError(input_name, "Input file could not be read")

    if compiler_options.verbose_operators:
        nng.print_operators()

    if compiler_options.timing:
        stop = time.time()
        print("Model reading took %f s" % (stop - start))
        start = time.time()

    compiler_driver.compiler_driver(nng, arch, compiler_options, scheduler_options, network_type, output_basename)

    summary_csv_file = "{0}_summary_{1}.csv".format(output_basename, arch.system_config)
    stats_writer.write_summary_metrics_csv(nng, summary_csv_file, arch)

    stats_writer.print_performance_metrics(
        nng,
        show_cpu_operations=compiler_options.show_cpu_operations,
        verbose_weights=compiler_options.verbose_weights,
        verbose_cycle_estimate=compiler_options.verbose_cycle_estimate,
        arch=arch,
    )

    output_tfl_filename = output_basename + "_vela.tflite"
    if output_format == "tflite":
        tflite_writer.write_tflite(nng, output_tfl_filename)
    elif output_format == "raw":
        rawdata_writer.write_rawdata_output(nng, arch, output_basename)
    else:
        assert False, f"Unsupported output_format = {output_format}"

    if enable_debug_db:
        file_offsets = calculate_operator_file_offsets(output_tfl_filename)
        for idx, offset in enumerate(sorted(file_offsets)):
            sg = find_subgraph_with_command_stream_order(nng, idx)
            if sg is not None:
                DebugDatabase.set_stream_offset(sg, offset)
        debug_filename = output_basename + "_debug.xml"
        DebugDatabase.write(debug_filename, input_name, output_tfl_filename)

    if compiler_options.timing:
        stop = time.time()
        print("Compiler driver took %f s" % (stop - start))

    return nng


def process_regor(
    arch,
    input_name,
    accelerator,
    system_config,
    options,
    enable_debug_db,
    output_dir,
    output_format,
    verbose_weights=False,
    verbose_cycle_estimate=False,
    show_cpu_operations=False,
    verbose_performance=False,
):
    os.makedirs(output_dir, exist_ok=True)

    with open(input_name, "rb") as f:
        with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ) as network:
            fmt = get_format(network)
            compiled_model = regor.compile(accelerator, network, fmt, system_config, options=options, verbose=True)

    model_name = os.path.splitext(os.path.basename(input_name))[0]

    output_basename = os.path.join(output_dir, model_name)

    if output_format == "tflite":
        assert isinstance(compiled_model, regor.CompiledTFLiteModel)
        output_name = output_basename + "_vela.tflite"
        with open(output_name, "wb") as f:
            f.write(compiled_model.model)
    elif output_format == "raw":
        assert isinstance(compiled_model, regor.CompiledRawModel)
        rawdata_writer.write_rawdata_output_from_model(output_basename, compiled_model)
    else:
        assert False, f"Unsupported output_format = {output_format}"

    summary_csv_file = "{0}_summary_{1}.csv".format(output_basename, arch.system_config)

    if verbose_performance:
        stats_writer.write_regor_perlayer_performance_csv(
            arch, compiled_model.opt_database, compiled_model.perf_report, output_basename
        )
        stats_writer.print_regor_perlayer_performance(
            arch, compiled_model.opt_database, compiled_model.perf_report, output_basename
        )

    stats_writer.print_regor_performance_metrics(
        arch,
        compiled_model.perf_report,
        model_name,
        summary_csv_file,
        compiled_model.opt_database,
        verbose_weights,
        verbose_cycle_estimate,
        show_cpu_operations,
    )
    if enable_debug_db:
        stats_writer.write_regor_db(compiled_model.opt_database, output_basename)


def find_subgraph_with_command_stream_order(nng, idx):
    for sg in nng.subgraphs:
        if sg.generated_stream_id == idx:
            return sg
    return None


def calculate_operator_file_offsets(name: str):
    # Read the vela optimized TFLite file
    with open(name, "rb") as f:
        buf = bytearray(f.read())
    # Calculate the file offsets for each custom operator
    file_offsets = []
    model = Model.GetRootAsModel(buf, 0)
    for idx in range(model.SubgraphsLength()):  # However only one subgraph is supported as of now
        sg = model.Subgraphs(idx)
        for idx in range(sg.OperatorsLength()):
            operator = sg.Operators(idx)
            if model.OperatorCodes(operator.OpcodeIndex()).CustomCode() is not None:
                tensor_idx = operator.Inputs(0)
                tensor = sg.Tensors(tensor_idx)
                buffer = model.Buffers(tensor.Buffer())
                offset = flatbuffers.number_types.UOffsetTFlags.py_type(buffer._tab.Offset(4))
                file_offsets.append(buffer._tab.Vector(offset))
    return file_offsets


def print_subgraph_io_summary(nng):
    """Print a summary of all the input and output tensor sizes for all subgraphs.
    Also displays the total tensor size and the memory used area for sram.
    """

    print("Subgraph IO Summary")
    print("-------------------")
    print(f"NNG: {nng.name}")
    max_sg_size = 0
    for sg in reversed(nng.subgraphs):
        print(f"  NNG Subgraph: {sg.name} = {sg.placement}")
        sg_size = 0

        if hasattr(sg, "scratch_tensor") and sg.scratch_tensor is not None:
            sg_tensors = sg.input_tensors + [sg.scratch_tensor] + sg.output_tensors
        else:
            sg_tensors = sg.input_tensors + sg.output_tensors

        for tens in sg_tensors:
            if tens in sg.input_tensors:
                tens_dir = "In"
            elif tens in sg.output_tensors:
                tens_dir = "Out"
            else:
                tens_dir = "In/Out"

            size = tens.elements() * tens.element_size() / 1024.0
            sg_size = sg_size + size
            print(f"         Tensor [{tens_dir}]: {tens.name} = {size} KiB")

        print(f"      Total Size = {sg_size} KiB")
        print(f"      SRAM Memory Used = {sg.memory_used.get(MemArea.Sram, 0) / 1024.0} KiB")
        max_sg_size = max(sg_size, max_sg_size)

    print(f"   Maximum NNG Subgraph Size = {max_sg_size} KiB")


def generate_supported_ops():
    # Exclude network type from generation by adding value to exclude list.
    # To easily exclude NetworkType from generated documentation.
    exclude_generation_network_type_value = [NetworkType.TOSA.value]

    def _exclude_list_names(constraint, exclude_list):
        constraints_excluded_names = [
            optype_to_builtintype(op) for op, exclude_constraint in exclude_list if constraint in exclude_constraint
        ]
        return f" - [{', '.join(sorted(constraints_excluded_names))}]" if constraints_excluded_names else ""

    # Add license for supported ops
    lines = [
        "<!--",
        "SPDX-FileCopyrightText: Copyright 2020-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>",
        "",
        "SPDX-License-Identifier: Apache-2.0",
        "",
        "Licensed under the Apache License, Version 2.0 (the License); you may",
        "not use this file except in compliance with the License.",
        "You may obtain a copy of the License at",
        "",
        "www.apache.org/licenses/LICENSE-2.0",
        "",
        "Unless required by applicable law or agreed to in writing, software",
        "distributed under the License is distributed on an AS IS BASIS, WITHOUT",
        "WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.",
        "See the License for the specific language governing permissions and",
        "limitations under the License.",
        "-->",
        "",
    ]

    lines += [
        "# Supported Ops",
        "",
        "This file was automatically generated by Vela using the `--supported-ops-report` parameter.  ",
        f"Vela version: `{__version__}`",
        "",
        "This file complies with",
        "[**Gitiles Markdown syntax**](https://gerrit.googlesource.com/gitiles/+/HEAD/Documentation/markdown.md)",
        "",
        "Summary table of constraints for:",
    ]

    # Ethos-U55 and Ethos-U65 TFLite and TOSA
    for network_type in NetworkType:
        if network_type.value in exclude_generation_network_type_value:
            continue

        lines += [
            f"- [Ethos-U55 and Ethos-U65 {network_type.name}]"
            f"(#ethos-u55-and-ethos-u65-{network_type.name.lower()}-summary-table)",
        ]

    # Ethos-U85 TFLite
    lines += [
        "- [Ethos-U85 TFLite](#ethos-u85-tflite-summary-table)",
    ]

    # Ethos-U55 and Ethos-U65 TFLite and TOSA
    for network_type in NetworkType:
        if network_type.value in exclude_generation_network_type_value:
            continue

        lines += [
            "",
            f"## Ethos-U55 and Ethos-U65 {network_type.name} Summary Table",
            "",
        ]
        if network_type == NetworkType.TFLite:
            lines += [
                "The table below contains TFLite operators that can be placed on the Ethos-U55 and Ethos-U65.  ",
                "If the constraints are not met, then that operator will be scheduled on the CPU instead.  ",
                "For any other TFLite operator not listed, will be left untouched and scheduled on the CPU.  ",
                "Please check the supported operator list for your chosen runtime for further information.",
                "",
                "| Operator | TFLite Constraints |",
                "| --- | --- |",
            ]
            semantic_checker = TFLiteSemantic()
            supported = TFLiteSupportedOperators()
        elif network_type == NetworkType.TOSA:
            lines += [
                "The table below contains TOSA operators that can be placed on the Ethos-U NPU.  ",
                "Note: There is limited support for compiling a TOSA neural network (EXPERIMENTAL).  ",
                "The related constraints have not yet been populated in the list.",
                "",
                "| Operator | TOSA Constraints |",
                "| --- | --- |",
            ]
            semantic_checker = TosaSemantic()
            supported = TosaSupportedOperators()
        else:
            raise ValueError

        op_constraint_links = []
        op_list = sorted(((op, builtin_operator_name_map[op]) for op in builtin_operator_map), key=lambda x: x[1])
        for op, name in op_list:
            internal_op = builtin_operator_map[op][0]
            if internal_op in TFLiteSupportedOperators.supported_operators:
                links = f"[Generic](#{network_type.name.lower()}-generic-constraints)"
                if (
                    internal_op in supported.specific_constraints
                    or internal_op in semantic_checker.specific_constraints
                ):
                    links += (
                        f", [Specific](#ethos-u55-and-ethos-u65-{network_type.name.lower()}-{name.lower()}-constraints)"
                    )
                    op_constraint_links.append((internal_op, name))
                lines.append(f"| {name} | {links} |")
        # Ethos-U85 TFLite
        lines += [
            """
## Ethos-U85 TFLite Summary Table

The table below contains TFLite operators that can be placed on the Ethos-U85.
If the constraints are not met, then that operator will be scheduled on the CPU instead.
For any other TFLite operator not listed, will be left untouched and scheduled on the CPU.
Please check the supported operator list for your chosen runtime for further information.

| Operator | TFLite Constraints |
| --- | --- |
| ABS | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-abs-constraints) |
| ADD | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-add-constraints) |
| ARG_MAX | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-arg_max-constraints) |
| AVERAGE_POOL_2D | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-average_pool_2d-constraints) |
| BATCH_MATMUL | [Generic](#tflite-generic-constraints) |
| CONCATENATION | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-concatenation-constraints) |
| CONV_2D | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-conv_2d-constraints) |
| DEPTHWISE_CONV_2D | [Generic](#tflite-generic-constraints), \
[Specific](#ethos-u85-tflite-depthwise_conv_2d-constraints) |
| EXP | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-exp-constraints) |
| EXPAND_DIMS | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-expand_dims-constraints) |
| FULLY_CONNECTED | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-fully_connected-constraints) |
| GATHER | [Generic](#tflite-generic-constraints) | [Specific](#ethos-u85-tflite-gather-constraints) |
| HARD_SWISH | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-hard_swish-constraints) |
| LEAKY_RELU | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-leaky_relu-constraints) |
| LOG | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-log-constraints) |
| LOGISTIC | [Generic](#tflite-generic-constraints) |
| MAXIMUM | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-maximum-constraints) |
| MAX_POOL_2D | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-max_pool_2d-constraints) |
| MEAN | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-mean-constraints) |
| MINIMUM | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-minimum-constraints) |
| MIRROR_PAD | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-mirror_pad-constraints) |
| MUL | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-mul-constraints) |
| PACK | [Generic](#tflite-generic-constraints) |
| PAD | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-pad-constraints) |
| PRELU | [Generic](#tflite-generic-constraints) |
| QUANTIZE | [Generic](#tflite-generic-constraints) |
| RELU | [Generic](#tflite-generic-constraints) |
| RELU6 | [Generic](#tflite-generic-constraints) |
| RELU_0_TO_1 | [Generic](#tflite-generic-constraints) |
| RELU_N1_TO_1 | [Generic](#tflite-generic-constraints) |
| RESHAPE | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-reshape-constraints) |
| RESIZE_BILINEAR | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-resize_bilinear-constraints) |
| RESIZE_NEAREST_NEIGHBOR | [Generic](#tflite-generic-constraints), \
[Specific](#ethos-u85-tflite-resize_nearest_neighbor-constraints) |
| RSQRT | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-rsqrt-constraints) |
| SCATTER | [Generic](#tflite-generic-constraints) | [Specific](#ethos-u85-tflite-scatter-constraints) |
| SELECT | [Generic](#tflite-generic-constraints) |
| SELECT_V2 | [Generic](#tflite-generic-constraints) |
| SHAPE | [Generic](#tflite-generic-constraints) |
| SLICE | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-slice-constraints) |
| SOFTMAX | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-softmax-constraints) |
| SPLIT | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-split-constraints) |
| SPLIT_V | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-split_v-constraints) |
| SQUARED_DIFFERENCE | [Generic](tflite-generic-constraints), \
[Specific](#ethos-u85-tflite-squared_difference-constraints) |
| SQUEEZE | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-squeeze-constraints) |
| STRIDED_SLICE | [Generic](#tflite-generic-constraints), \
[Specific](#ethos-u85-tflite-strided_slice-constraints) |
| SUB | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-sub-constraints) |
| TANH | [Generic](#tflite-generic-constraints) |
| TRANSPOSE | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-transpose-constraints) |
| TRANSPOSE_CONV | [Generic](#tflite-generic-constraints), [Specific](#ethos-u85-tflite-transpose_conv-constraints) |
| UNIDIRECTIONAL_SEQUENCE_LSTM | [Generic](#tflite-generic-constraints), \
[Specific](#ethos-u85-tflite-unidirectional_sequence_lstm-constraints) |
| UNPACK | [Generic](#tflite-generic-constraints) |
"""
        ]
        lines += [
            "",
            f"### {network_type.name} Generic Constraints",
            "",
            "This is a list of constraints that most operators must satisfy in order to be scheduled on the NPU.",
            "(Operators excluded from certain constraints are shown in brackets [ ] )\n" "",
        ]
        for constraint in semantic_checker.generic_constraints:
            # Markdown needs two spaces at the end of a line to render it as a separate line
            reason = constraint.__doc__.replace("\n", "  \n")
            exclude_list = TFLiteSemantic.get_generic_constraint_exclude_list().items()
            lines.append(f"- {reason}{_exclude_list_names(constraint, exclude_list)}")
        for constraint in supported.generic_constraints:
            # Markdown needs two spaces at the end of a line to render it as a separate line
            reason = constraint.__doc__.replace("\n", "  \n")
            exclude_list = supported.generic_constraints_exceptions.items()
            lines.append(f"- {reason}{_exclude_list_names(constraint, exclude_list)}")
        lines += ["", "## Ethos-U55 and Ethos-U65 Specific Operator constraints"]
        for op, name in op_constraint_links:
            lines += [
                "",
                f"### Ethos-U55 and Ethos-U65 {network_type.name} {name} Constraints",
                "",
                f"This is a list of constraints that the {name} operator must satisfy in order to be scheduled on the"
                " NPU.",
                "",
            ]
            for constraint in semantic_checker.specific_constraints[op]:
                # Markdown needs two spaces at the end of a line to render it as a separate line
                reason = constraint.__doc__.replace("\n", "  \n")
                lines.append(f"- {reason}")
            for constraint in supported.specific_constraints[op]:
                # Markdown needs two spaces at the end of a line to render it as a separate line
                reason = constraint.__doc__.replace("\n", "  \n")
                lines.append(f"- {reason}")

    # Ethos-U85 TFLite
    lines += [
        """
## Ethos-U85 Specific Operator Constraints

### Ethos-U85 TFLite ABS Constraints

This is a list of constraints that the ABS operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match

### Ethos-U85 TFLite ADD Constraints

This is a list of constraints that the ADD operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- Both Input data types must match
- For IFM that are signed, OFM must also be signed
- For IFM that are unsigned, OFM must either be the same type or int32
- Broadcasting is only allowed for rank indices with dimension 1, from either IFM1 or IFM2

### Ethos-U85 TFLite ARG_MAX Constraints

This is a list of constraints that the ARG_MAX operator must satisfy in order to be scheduled on the NPU.

- OFM must be int32 or int64

### Ethos-U85 TFLite AVERAGE_POOL_2D Constraints

This is a list of constraints that the AVERAGE_POOL_2D operator must satisfy in order to be scheduled on the NPU.

- Stride values for both width and height must be integer types
- IFM and OFM data types must match
- Kernel filter values for both width and height must be integer types
- Stride values for both width and height must be in the range [1, 3]
- Kernel filter values for both width and height must be in the range [1, 8]
- VALID padding: Kernel filter height must be in the range [1, 256]
- VALID padding: Product of kernel filter width and height must be in the range [1, 65536]

### Ethos-U85 TFLite CONCATENATION Constraints

This is a list of constraints that the CONCATENATION operator must satisfy in order to be scheduled on the NPU.

- Axis attribute must exist
- Axis attribute must be in the range [0, <ofm_dimensions>)
- All Input dimensionalities must match OFM dimensionality
- All Input dimensions must match OFM dimension in all axes except the one defined by the axis attribute
- The size of the OFM axis must match the sum of all IFM axis defined by the axis attribute

### Ethos-U85 TFLite CONV_2D Constraints

This is a list of constraints that the CONV_2D operator must satisfy in order to be scheduled on the NPU.

- Stride values for both width and height must be integer types
- IFM depth must be a whole multiple of the filter kernel depth
- Dilation factor values for both width and height must be integer types
- Stride values for both width and height must be in the range [1, 3]
- Dilated kernel height must be in the range [1, 64]
- Product of dilated kernel width and height must be in the range [1, 4096]
- Weight tensor must be 8-bit
- Weight tensor must be constant
- The sum of the weights cannot exceed 8323072
- Optional Bias tensor must be of shape: 1D
- Optional Bias tensor must be of type: int32, int64
- Optional Bias tensor values must fit within 40-bits

### Ethos-U85 TFLite DEPTHWISE_CONV_2D Constraints

This is a list of constraints that the DEPTHWISE_CONV_2D operator must satisfy in order to be scheduled on the NPU.

- Stride values for both width and height must be integer types
- Dilation factor values for both width and height must be integer types
- Dilated kernel height must be in the range [1, 64]
- Product of dilated kernel width and height must be in the range [1, 4096]
- Weight tensor must be 8-bit
- Weight tensor must be constant
- The sum of the weights cannot exceed 8323072
- Optional Bias tensor must be of shape: 1D
- Optional Bias tensor must be of type: int32, int64
- Optional Bias tensor values must fit within 40-bits
- Stride values for both width and height must be between 1 and 3
- For depth multipliers > 1, IFM channels must be 1 and OFM channels must be equal to the depth multiplier

### Ethos-U85 TFLite EXP Constraints

This is a list of constraints that the EXP operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match
- IFM must be int8 or int16

### Ethos-U85 TFLite EXPAND_DIMS Constraints

This is a list of constraints that the EXPAND_DIMS operator must satisfy in order to be scheduled on the NPU.

- Input and output quantisation must match.
- Input and output number of elements must match.

### Ethos-U85 TFLite FULLY_CONNECTED Constraints

This is a list of constraints that the FULLY_CONNECTED operator must satisfy in order to be scheduled on the NPU.

- The output tensor(s) must have 2D shape
- The IFM and OFM must have the same number of dimensions if keep_num_dims is set to true
- Weight tensor must be 8-bit
- Weight tensor must be constant
- Optional Bias tensor must be of shape: 1D
- Optional Bias tensor must be of type: int32, int64
- Optional Bias tensor values must fit within 40-bits

### Ethos-U85 TFLite GATHER Constraints

This is a list of constraints that the GATHER operator must satisfy in order to be scheduled on the NPU.

- Axis parameter must equal batch dim parameter

### Ethos-U85 TFLite HARD_SWISH Constraints

This is a list of constraints that the HARD_SWISH operator must satisfy in order to be scheduled on the NPU.

- IFM must be int8 or uint8
- IFM and OFM data types must match

### Ethos-U85 TFLite LEAKY_RELU Constraints

This is a list of constraints that the LEAKY_RELU operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match

### Ethos-U85 TFLite MAXIMUM Constraints

This is a list of constraints that the MAXIMUM operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match
- Both Input quantization parameters must match OFM quantization parameters
- Broadcasting is only allowed for rank indices with dimension 1, from either IFM1 or IFM2

### Ethos-U85 TFLite LOG Constraints

This is a list of constraints that the LOG operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match
- IFM must be int8 or int16

### Ethos-U85 TFLite MAX_POOL_2D Constraints

This is a list of constraints that the MAX_POOL_2D operator must satisfy in order to be scheduled on the NPU.

- Stride values for both width and height must be integer types
- IFM and OFM data types must match
- Kernel filter values for both width and height must be integer types
- Stride values for both width and height must be in the range [1, 3]
- Kernel filter height must be in the range [1, 256]
- Product of kernel filter width and height must be in the range [1, 65536]

### Ethos-U85 TFLite MEAN Constraints

This is a list of constraints that the MEAN operator must satisfy in order to be scheduled on the NPU.

- Input tensor must be at least 2D
- Requirements for axis parameter:  
        When IFM tensor is 2D:  
          - Reduction in both axes is supported.  
        When IFM tensor is 3D or 4D:  
          - Reduction in Batch axis is only supported if batch size is 1.  
          - Reduction in both Height and Width axes is supported.  
          - Reduction in Depth axis is supported if at least one of H,W,C are of size 1.
- Product of reduced axes must be no greater than:  
        - 16777216 for signed 8-bit inputs.  
        - 8388608 for unsigned 8-bit inputs.  
        - 65536 for signed 16-bit inputs.
- If Width axis is reduced its shape must be no greater than 4096.
- If Depth axis is reduced its shape must be no greater than 4096.

### Ethos-U85 TFLite MINIMUM Constraints

This is a list of constraints that the MINIMUM operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match
- Both Input quantization parameters must match OFM quantization parameters
- Broadcasting is only allowed for rank indices with dimension 1, from either IFM1 or IFM2

### Ethos-U85 TFLite MIRROR_PAD Constraints

This is a list of constraints that the MIRROR_PAD operator must satisfy in order to be scheduled on the NPU.

- The padding tensor must have the shape [3,2] or [4,2]
- The pad tensor can only pad width and height
- Pad tensor must be of type: int32, int64
- The number of pad values for each direction must not be larger than the ifm size in that dimension

### Ethos-U85 TFLite MUL Constraints

This is a list of constraints that the MUL operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- Both Input data types must match
- For IFM that are signed, OFM must also be signed
- For IFM that are unsigned, OFM must either be the same type or int32
- Broadcasting is only allowed for rank indices with dimension 1, from either IFM1 or IFM2

### Ethos-U85 TFLite PAD Constraints

This is a list of constraints that the PAD operator must satisfy in order to be scheduled on the NPU.

- Number of input tensors must be exactly 2
- The padding tensor must be constant
- Shape of output tensor must equal to size of input tensor plus padding
- The padding tensor must have the shape [3,2] or [4,2]
- The pad tensor can only pad width and height
- Pad tensor must be of type: int32, int64

### Ethos-U85 TFLite RESHAPE Constraints

This is a list of constraints that the RESHAPE operator must satisfy in order to be scheduled on the NPU.

- Input and output quantisation must match.
- Input and output number of elements must match.
- Shape must be constant

### Ethos-U85 TFLite RESIZE_BILINEAR Constraints

This is a list of constraints that the RESIZE_BILINEAR operator must satisfy in order to be scheduled on the NPU.

- The size tensor must match the output tensor shape
- Both align_corners and half_pixel_centers can't be True

### Ethos-U85 TFLite RESIZE_NEAREST_NEIGHBOR Constraints

This is a list of constraints that the RESIZE_NEAREST_NEIGHBOR operator must satisfy in order to be scheduled on the \
NPU.

- The size tensor must match the output tensor shape
- Both align_corners and half_pixel_centers can't be True

### Ethos-U85 TFLite RSQRT Constraints

This is a list of constraints that the RSQRT operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- IFM and OFM data types must match
- IFM must be int8

### Ethos-U85 TFLite SCATTER Constraints

This is a list of constraints that the SCATTER operator must satisfy in order to be scheduled on the NPU.

- Index tensor must be constant
- Index tensor must not contain any duplicate values
- Last dimension of index tensor must be 1
- Shape tensor must be constant

### Ethos-U85 TFLite SLICE Constraints

This is a list of constraints that the SLICE operator must satisfy in order to be scheduled on the NPU.

- Begin and Size Input tensors must be constant

### Ethos-U85 TFLite SOFTMAX Constraints

This is a list of constraints that the SOFTMAX operator must satisfy in order to be scheduled on the NPU.

- IFM and OFM shapes must match
- IFM and OFM data types must match
- Beta value needs to be positive

### Ethos-U85 TFLite SPLIT Constraints

This is a list of constraints that the SPLIT operator must satisfy in order to be scheduled on the NPU.

- Axis value must be in the range [-RANK(IFM) to +RANK(IFM))
- Axis must be divisible by number of splits

### Ethos-U85 TFLite SPLIT_V Constraints

This is a list of constraints that the SPLIT_V operator must satisfy in order to be scheduled on the NPU.

- Only one size is allowed to be inferred

### Ethos-U85 TFLite SQUARED_DIFFERENCE Constraints

This is a list of constraints that the SQUARED_DIFFERENCE operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape

### Ethos-U85 TFLite SQUEEZE Constraints

This is a list of constraints that the SQUEEZE operator must satisfy in order to be scheduled on the NPU.

- Input and output quantisation must match.
- Input and output number of elements must match.

### Ethos-U85 TFLite STRIDED_SLICE Constraints

This is a list of constraints that the STRIDED_SLICE operator must satisfy in order to be scheduled on the NPU.

- Exactly 4 Input tensors are required
- Begin, End and Stride Input tensors must be constant
- ellipsis_mask must be 0
- new_axis_mask and shrink_axis_mask cannot both be set
- Slice 'end' values must be greater than 'begin' values
- All Strides values must be 1
- Offset attribute must be False

### Ethos-U85 TFLite SUB Constraints

This is a list of constraints that the SUB operator must satisfy in order to be scheduled on the NPU.

- At least one Input's shape must match the OFM's shape
- Both Input data types must match
- For IFM that are signed, OFM must also be signed
- For IFM that are unsigned, OFM must either be the same type or int32
- Broadcasting is only allowed for rank indices with dimension 1, from either IFM1 or IFM2

### Ethos-U85 TFLite TRANSPOSE Constraints

This is a list of constraints that the TRANSPOSE operator must satisfy in order to be scheduled on the NPU.

- Permutation array must be a 1D tensor with RANK(IFM) elements
- Permutation array must have constant values in the range [0, RANK(IFM))

### Ethos-U85 TFLite TRANSPOSE_CONV Constraints

This is a list of constraints that the TRANSPOSE_CONV operator must satisfy in order to be scheduled on the NPU.

- Stride values for both width and height must be integer types
- Dilated kernel height must be in the range [1, 64]
- Product of dilated kernel width and height must be in the range [1, 4096]
- Weight tensor must be 8-bit
- Weight tensor must be constant
- The sum of the weights cannot exceed 8323072
- Optional Bias tensor must be of shape: 1D
- Optional Bias tensor must be of type: int32, int64
- Optional Bias tensor values must fit within 40-bits
- Stride values for width and height must match one of the following criteria:  
        Stride values WxH must be 1x1 or 2x2  
        Stride WxH 2x1 supported if ifm height and kernel height = 1
- SAME padding: OFM dimensions must equal IFM dimensions multiplied by stride
- VALID padding: OFM dimensions must equal IFM dimensions multiplied by stride,  
        minus difference between kernel size and stride

### Ethos-U85 TFLite UNIDIRECTIONAL_SEQUENCE_LSTM Constraints

This is a list of constraints that the UNIDIRECTIONAL_SEQUENCE_LSTM operator must satisfy in order to be scheduled on \
the NPU.

- IFM must be int8 or int16
- IFM and OFM data types must match
- IFM and OFM must have 3D shape
- Must have 24 input tensors
- Must have 5 intermediate tensors
- State tensors must be variable
- Must not use CIFG
- Must not use Peephole
- Must not use Projection
- Must not use Normalisation
- All input and recurrent weights must be available
- All recurrent weights must be 2D
"""
    ]

    # Note. this will generate the file in the CWD
    filepath = os.path.join(os.getcwd(), "SUPPORTED_OPS.md")
    with open(filepath, "wt") as md:
        md.writelines(line + "\n" for line in lines)
        print(f"Report file: {filepath}")


def get_compiler_config(
    enable_debug_db: bool,
    verbose_all: bool,
    verbose_high_level_command_stream: bool,
    verbose_register_command_stream: bool,
    optimize: str,
    arena_cache_size: int,
    verbose_schedule: bool,
    verbose_allocation: bool,
    verbose_graph: bool,
    verbose_quantization: bool,
    verbose_packing: bool,
    verbose_performance: bool,
    show_cpu_operations: bool,
    output_format: str,
    disable_chaining: bool,
    disable_fwd: bool,
    disable_cascading: bool,
    disable_buffering: bool,
    cop_format: str,
    separate_io_regions: bool,
    cpu_tensor_alignment: int,
) -> str:
    """Build compiler config file."""
    config = "\n[compiler]\n"
    if verbose_high_level_command_stream:
        config += "verbose_high_level_command_stream=true\n"
    if verbose_register_command_stream:
        config += "verbose_register_command_stream=true\n"
    if verbose_performance or show_cpu_operations or enable_debug_db:
        config += "enable_db=true\n"
    if output_format == "raw":
        config += "output_format=Raw\n"
    else:
        config += "output_format=TFLite\n"
    config += f"cop_format={cop_format}\n"

    config += "\n[scheduler]\n"
    config += f"optimize={optimize}\n"
    config += f"arena_size_limit={arena_cache_size}\n"
    if verbose_schedule:
        config += "verbose=true\n"
    if verbose_allocation:
        config += "verbose_allocation=true\n"
    config += "disable_feature="
    if disable_chaining:
        config += "Grouping|"
    if disable_fwd:
        config += "FWD|"
    if disable_cascading:
        config += "Cascading|"
    if disable_buffering:
        config += "WeightBuffering|"
    config = config.rstrip("|") + "\n"
    if separate_io_regions:
        config += "separate_io_regions=true\n"
    config += f"cpu_tensor_alignment={cpu_tensor_alignment}\n"

    config += "\n[graph]\n"
    if verbose_graph:
        config += "verbose=true\n"
    if verbose_quantization:
        config += "verbose_quantization=true\n"

    return config


def get_format(in_data: mmap.mmap) -> str:
    """Infere format based on input file."""
    ret = "UNDEFINED"
    if in_data.size() < 8:
        return ret
    second_word = int.from_bytes(in_data[4:8], "little")
    if second_word == TFLITE_MAGIC:
        ret = "TFLITE"
    if second_word == TOSA_MAGIC:
        ret = "TOSA"
    return ret


def list_config_files():
    print("Available config files:")
    path_length = len(architecture_features.CONFIG_FILES_PATH + os.path.sep)
    for config in glob.glob(os.path.join(architecture_features.CONFIG_FILES_PATH, "*", "*.ini")):
        print(config[path_length:])


def list_configs(config_filename):
    vela_config_filename = config_filename
    if not os.path.isfile(vela_config_filename):
        vela_config_filename = os.path.join(architecture_features.CONFIG_FILES_PATH, config_filename)
        if not os.path.isfile(vela_config_filename):
            assert False, f"Cannot find config file {config_filename}"

    if os.path.splitext(vela_config_filename)[1] != ".ini":
        assert False, f"Specified file {config_filename} is not a Vela config file"

    print(f"Configurations defined in {vela_config_filename}:")
    vela_config = ConfigParser()
    vela_config.read(vela_config_filename)
    for section in vela_config.sections():
        print(f"   {section}")


def main(argv: Optional[List[str]] = None) -> int:
    """Run the main entry point."""
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(prog="vela", description="Neural network model compiler for Arm Ethos-U NPUs")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--api-version", action="version", version=API_VERSION, help="Displays the version of the external API."
    )
    parser.add_argument(
        "--supported-ops-report",
        action="store_true",
        help="Generate the SUPPORTED_OPS.md file in the current working directory and exit",
    )

    parser.add_argument(
        "--list-config-files",
        action="store_true",
        help=(
            "Display all available configurations in the `config_files` folder and exit. To select config file, "
            "use the --config argument with one of the listed config files (For example: --config Arm/vela.ini )"
        ),
    )
    parser.add_argument(
        "--list-configs",
        type=str,
        help=("Display all configurations defined in the specified config file"),
    )

    # set network nargs to be optional to allow the support-ops-report CLI option to be used standalone
    parser.add_argument(
        "network",
        metavar="NETWORK",
        type=str,
        nargs="?",
        default=None,
        help="Filename of the input network",
    )
    parser.add_argument(
        "--output-dir", type=str, default="output", help="Output directory to write files to (default: %(default)s)"
    )
    parser.add_argument(
        "--output-format",
        type=str,
        default="tflite",
        choices=["tflite", "raw"],
        help="Output format (default: %(default)s)",
    )
    parser.add_argument(
        "--enable-debug-db",
        action="store_true",
        default=None,
        help="Enables the calculation and writing of a network debug database to output directory",
    )
    parser.add_argument(
        "--config",
        type=str,
        action="append",
        help="Vela configuration file(s) in Python ConfigParser .ini file format",
    )
    parser.add_argument("--verbose-all", action="store_true", help="Enable all verbose options")
    parser.add_argument(
        "--verbose-config", action="store_true", help="Enable system configuration and memory mode debug"
    )
    parser.add_argument("--verbose-graph", action="store_true", help="Enable graph optimizer debug")
    parser.add_argument("--verbose-quantization", action="store_true", help="Enable quantization debug")
    parser.add_argument("--verbose-packing", action="store_true", help="Enable pass packing debug")
    parser.add_argument("--verbose-tensor-purpose", action="store_true", help="Enable tensor purpose debug")
    parser.add_argument("--verbose-tensor-format", action="store_true", help="Enable tensor format debug")
    parser.add_argument("--verbose-schedule", action="store_true", help="Enable schedule debug")
    parser.add_argument("--verbose-allocation", action="store_true", help="Enable tensor allocation debug")
    parser.add_argument(
        "--verbose-high-level-command-stream", action="store_true", help="Enable high level command stream debug"
    )
    parser.add_argument(
        "--verbose-register-command-stream", action="store_true", help="Enable register command stream debug"
    )
    parser.add_argument("--verbose-operators", action="store_true", help="Enable operator list debug")
    parser.add_argument("--verbose-weights", action="store_true", help="Enable weights information debug")
    parser.add_argument("--verbose-cycle-estimate", action="store_true", help="Enable cycle estimate information debug")
    parser.add_argument("--verbose-performance", action="store_true", help="Enable performance information debug")
    parser.add_argument("--verbose-progress", action="store_true", help="Enable progress information debug")
    parser.add_argument(
        "--show-cpu-operations", action="store_true", help="Show the operations that fall back to the CPU"
    )
    parser.add_argument("--timing", action="store_true", help="Time the compiler doing operations")
    parser.add_argument(
        "--force-symmetric-int-weights",
        action="store_true",
        help="Forces all zero points to 0 for signed integer weights",
    )
    parser.add_argument(
        "--accelerator-config",
        type=str,
        default="ethos-u55-256",
        choices=list(architecture_features.Accelerator.member_list()),
        help="Accelerator configuration to use (default: %(default)s)",
    )
    parser.add_argument(
        "--system-config",
        type=str,
        default=architecture_features.ArchitectureFeatures.DEFAULT_CONFIG,
        help="System configuration to select from the Vela configuration file (default: %(default)s)",
    )
    parser.add_argument(
        "--memory-mode",
        type=str,
        default=architecture_features.ArchitectureFeatures.DEFAULT_CONFIG,
        help="Memory mode to select from the Vela configuration file (default: %(default)s)",
    )
    parser.add_argument(
        "--tensor-allocator",
        default=TensorAllocator.HillClimb,
        type=lambda s: TensorAllocator[s],
        choices=list(TensorAllocator),
        help="Tensor Allocator algorithm (default: %(default)s)",
    )
    parser.add_argument(
        "--show-subgraph-io-summary",
        action="store_true",
        help="Shows a summary of all the subgraphs and their inputs and outputs",
    )
    parser.add_argument(
        "--max-block-dependency",
        type=int,
        default=architecture_features.ArchitectureFeatures.MAX_BLOCKDEP,
        choices=range(0, architecture_features.ArchitectureFeatures.MAX_BLOCKDEP + 1),
        help=(
            "Set the maximum value that can be used for the block dependency between npu kernel operations"
            " (default: %(default)s)"
        ),
    )
    parser.add_argument(
        "--optimise",
        type=lambda s: scheduler.OptimizationStrategy[s],
        default=scheduler.OptimizationStrategy.Performance,
        choices=list(scheduler.OptimizationStrategy),
        help=(
            "Set the optimisation strategy. The Size strategy results in minimal SRAM usage (does not use"
            " arena-cache-size). The Performance strategy results in maximal performance (uses the arena-cache-size"
            " if specified) (default: %(default)s)"
        ),
    )
    parser.add_argument(
        "--arena-cache-size",
        type=int,
        help=(
            "Set the size of the arena cache memory area, in bytes. If specified, this option overrides the memory"
            " mode attribute with the same name in a Vela configuration file"
        ),
    )
    parser.add_argument(
        "--cpu-tensor-alignment",
        type=int,
        default=Tensor.AllocationQuantum,
        help=(
            "Controls the allocation byte alignment of CPU tensors including Ethos-U Custom"
            " operator inputs and outputs (default: %(default)s Bytes)"
        ),
    )
    parser.add_argument(
        "--recursion-limit",
        type=int,
        default=1000,
        help="Set the recursion depth limit, may result in RecursionError if too low (default: %(default)s)",
    )
    parser.add_argument(
        "--hillclimb-max-iterations",
        type=int,
        default=HillClimbAllocator.MAX_ITERATIONS,
        help="Set the maximum number of iterations the Hill Climb tensor allocator will run (default: %(default)s)",
    )
    parser.add_argument(
        "--cop-format",
        choices=["COP1", "COP2"],
        default="COP1",
    )
    parser.add_argument(
        "--separate-io-regions",
        action="store_true",
        help="Use separate regions for input and output tensors (implies COP2 driver actions format)",
    )

    # debug options
    parser.add_argument("--debug-force-regor", action="store_true", help="Debug: Force the use of the regor")
    parser.add_argument("--disable-chaining", action="store_true", default=False)
    parser.add_argument("--disable-fwd", action="store_true", default=False)
    parser.add_argument("--disable-cascading", action="store_true", default=False)
    parser.add_argument("--disable-buffering", action="store_true", default=False)

    args = parser.parse_args(argv)

    # Generate the supported ops report and exit
    if args.supported_ops_report:
        generate_supported_ops()
        return 0

    if args.list_config_files:
        list_config_files()
        return 0

    if args.list_configs:
        list_configs(args.list_configs)
        return 0

    if args.network is None:
        parser.error("the following argument is required: NETWORK")

    if args.cop_format == "COP1" and args.separate_io_regions:
        parser.error("Driver actions format 'COP2' is required for --separate-io-regions")

    def _parse_config(config):
        # Make sure the correct separator is used depending on OS
        config = os.path.normpath(config)

        if os.path.splitext(config)[1] != ".ini":
            raise InputFileError(config, "Configuration files must use the .ini extension")

        if (
            len(config.split(os.path.sep)) == 2
            and not config.startswith(os.path.sep)
            and not config.startswith(".")
            and not config.startswith("~")
        ):
            config_path = os.path.join(architecture_features.CONFIG_FILES_PATH, config)
        else:
            # Check if the configuration file is correctly placed inside the config_files directory
            if os.access(
                os.path.join(architecture_features.CONFIG_FILES_PATH, *config.split(os.path.sep)[-2:]),
                os.R_OK,
            ):
                rel_path = os.path.join(*config.split(os.path.sep)[-2:])
                print(
                    f"Warning: Consider accessing the configuration by --config {rel_path} since it is located "
                    "inside the config_files directory."
                )
            config_path = config

        if not os.access(config_path, os.R_OK):
            raise InputFileError(
                config_path,
                "File not found or is not readable. The configuration file is either not located in a folder "
                "directly under the `config_files` directory or its path has not been provided correctly.",
            )

        return config_path

    # check all config files exist because they will be read as a group
    config_files = [_parse_config(cfg) for cfg in args.config] if args.config else None

    if args.cpu_tensor_alignment < 16 or args.cpu_tensor_alignment & (args.cpu_tensor_alignment - 1) != 0:
        parser.error(
            "Invalid argument to --cpu-tensor-alignment = {} (must be greater than or equal to 16 and a power of 2)"
            "".format(args.cpu_tensor_alignment)
        )

    if args.verbose_all:
        for v in vars(args):
            if v.startswith("verbose") and v != "verbose_all":
                setattr(args, v, True)

    sys.setrecursionlimit(args.recursion_limit)

    arch = architecture_features.ArchitectureFeatures(
        vela_config_files=config_files,
        system_config=args.system_config,
        memory_mode=args.memory_mode,
        accelerator_config=args.accelerator_config,
        max_blockdep=args.max_block_dependency,
        verbose_config=args.verbose_config,
        arena_cache_size=args.arena_cache_size,
    )

    model_reader_options = model_reader.ModelReaderOptions()

    # The default behaviour to compile TFLite networks on Ethos-U55/U65 is to use Vela's Python compiler core (no name).
    # However, this can be overridden to use Vela's C++ compiler core (Regor) by using the --debug-force-regor option.
    # All Ethos-U85 or all TOSA network compilations use Vela's C++ compiler core (Regor).
    if arch.is_ethos_u85_system or args.network.lower().endswith(".tosa") or args.debug_force_regor:
        system_config = "[architecture]\n"
        system_config += f"macs={arch.num_macs_per_cycle}\n"
        system_config += f"cores={arch.ncores}\n"

        system_config += "[vela]\n"
        system_config += f"system_config_name={arch.system_config}\n"
        system_config += f"memory_mode_name={arch.memory_mode}\n"

        # TODO MLBEDSW-8190: Improve support for multiple config files
        for config_file in arch.vela_config_files:
            with open(config_file, "rb") as f:
                system_config += f.read().decode("utf-8")

        # Transform Vela accelerator config string format to Regor format
        accelerator_config = args.accelerator_config
        substrings = accelerator_config.split("-")
        substrings = [substring.capitalize() for substring in substrings[:-1]]
        accelerator = "".join(substrings)

        options = get_compiler_config(
            args.enable_debug_db,
            args.verbose_all,
            args.verbose_high_level_command_stream,
            args.verbose_register_command_stream,
            args.optimise,
            # TODO MLBEDSW-8191: Clean up arena_cache_size selection
            # arch.arena_cache_size will use value from CLI option if available, otherwise from config file
            arch.arena_cache_size,
            args.verbose_schedule,
            args.verbose_allocation,
            args.verbose_graph,
            args.verbose_quantization,
            args.verbose_packing,
            args.verbose_performance,
            args.show_cpu_operations,
            args.output_format,
            args.disable_chaining,
            args.disable_fwd,
            args.disable_cascading,
            args.disable_buffering,
            args.cop_format,
            args.separate_io_regions,
            args.cpu_tensor_alignment,
        )

        process_regor(
            arch,
            args.network,
            accelerator,
            system_config,
            options,
            args.enable_debug_db,
            args.output_dir,
            args.output_format,
            args.verbose_weights,
            args.verbose_cycle_estimate,
            args.show_cpu_operations,
            args.verbose_performance,
        )

    else:
        compiler_options = compiler_driver.CompilerOptions(
            verbose_graph=args.verbose_graph,
            verbose_quantization=args.verbose_quantization,
            verbose_packing=args.verbose_packing,
            verbose_tensor_purpose=args.verbose_tensor_purpose,
            verbose_tensor_format=args.verbose_tensor_format,
            verbose_allocation=args.verbose_allocation,
            verbose_high_level_command_stream=args.verbose_high_level_command_stream,
            verbose_register_command_stream=args.verbose_register_command_stream,
            verbose_operators=args.verbose_operators,
            verbose_weights=args.verbose_weights,
            verbose_cycle_estimate=args.verbose_cycle_estimate,
            verbose_performance=args.verbose_performance,
            verbose_progress=args.verbose_progress,
            show_cpu_operations=args.show_cpu_operations,
            tensor_allocator=args.tensor_allocator,
            timing=args.timing,
            force_symmetric_int_weights=args.force_symmetric_int_weights,
            output_dir=args.output_dir,
            cpu_tensor_alignment=args.cpu_tensor_alignment,
            hillclimb_max_iterations=args.hillclimb_max_iterations,
        )

        scheduler_options = scheduler.SchedulerOptions(
            optimization_strategy=args.optimise,
            sram_target=arch.arena_cache_size,
            verbose_schedule=args.verbose_schedule,
            verbose_progress=args.verbose_progress,
        )

        try:
            nng = process(
                args.network,
                args.enable_debug_db,
                arch,
                model_reader_options,
                compiler_options,
                scheduler_options,
                args.output_format,
            )

        except VelaError as e:
            print(e.data)
            return 1

        if args.show_subgraph_io_summary:
            print_subgraph_io_summary(nng)

    return 0



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Version(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Version()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsVersion(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def VersionBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # Version
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Version
    def _Major(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Version
    def _Minor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Version
    def _Patch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Version
    def _Draft(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

def VersionStart(builder):
    builder.StartObject(4)

def Start(builder):
    VersionStart(builder)

def VersionAdd_Major(builder, _Major):
    builder.PrependInt32Slot(0, _Major, -1)

def Add_Major(builder, _Major):
    VersionAdd_Major(builder, _Major)

def VersionAdd_Minor(builder, _Minor):
    builder.PrependInt32Slot(1, _Minor, -1)

def Add_Minor(builder, _Minor):
    VersionAdd_Minor(builder, _Minor)

def VersionAdd_Patch(builder, _Patch):
    builder.PrependInt32Slot(2, _Patch, -1)

def Add_Patch(builder, _Patch):
    VersionAdd_Patch(builder, _Patch)

def VersionAdd_Draft(builder, _Draft):
    builder.PrependBoolSlot(3, _Draft, 1)

def Add_Draft(builder, _Draft):
    VersionAdd_Draft(builder, _Draft)

def VersionEnd(builder):
    return builder.EndObject()

def End(builder):
    return VersionEnd(builder)



# SPDX-FileCopyrightText: Copyright 2021-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the License); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ethos-U-Vela package test."""
import os
import subprocess

import numpy as np


def test_ethos_u_vela_with_vela(tmp_path):
    """Testing ethos-u-vela with vela compiler core."""
    from network import TEST_NETWORK

    output_dir = tmp_path
    output_file = tmp_path / "model_vela.tflite"
    network_file = tmp_path / "model.tflite"
    network_file.write_bytes(TEST_NETWORK)
    cli = [
        "vela",
        "--accelerator-config",
        "ethos-u55-256",
        "--output-dir",
        output_dir,
        network_file,
    ]
    subprocess.check_call(cli)
    output = output_file.read_bytes()
    assert str.encode("COP1") in output
    assert str.encode("ethos-u") in output
    assert str.encode("_split_1_command_stream") in output


def test_ethos_u_vela_with_regor(tmp_path):
    """Testing ethos-u-vela with regor compiler core."""
    from network import TEST_NETWORK

    output_dir = tmp_path
    output_file = tmp_path / "model_vela.tflite"
    network_file = tmp_path / "model.tflite"
    network_file.write_bytes(TEST_NETWORK)
    cur_path = os.path.dirname(os.path.realpath(__file__))
    sc_path = os.path.join(cur_path, "..", "ethosu", "config_files", "Arm", "vela.ini")
    cli = [
        "vela",
        "--accelerator-config",
        "ethos-u85-512",
        "--config",
        sc_path,
        "--system-config",
        "Ethos_U85_SYS_DRAM_Mid",
        "--memory-mode",
        "Dedicated_Sram_384KB",
        "--output-dir",
        output_dir,
        network_file,
    ]
    subprocess.check_call(cli)
    output = output_file.read_bytes()
    assert output[4:8] == str.encode("TFL3")
    assert str.encode("COP1") in output
    assert str.encode("ethos-u") in output
    assert str.encode("ethos_u_command_stream") in output


def test_ethos_u_vela_with_regor_raw_output(tmp_path):
    """Testing ethos-u-vela with regor compiler core and raw output"""
    from network import TEST_NETWORK

    output_dir = tmp_path
    output_file = tmp_path / "model_vela.npz"
    network_file = tmp_path / "model.tflite"
    network_file.write_bytes(TEST_NETWORK)
    cur_path = os.path.dirname(os.path.realpath(__file__))
    sc_path = os.path.join(cur_path, "..", "ethosu", "config_files", "Arm", "vela.ini")
    cli = [
        "vela",
        "--accelerator-config",
        "ethos-u85-256",
        "--config",
        sc_path,
        "--system-config",
        "Ethos_U85_SYS_Flash_High",
        "--memory-mode",
        "Shared_Sram",
        "--output-dir",
        output_dir,
        "--output-format",
        "raw",
        network_file,
    ]
    subprocess.check_call(cli)
    raw = np.load(output_file)
    assert len(raw["cmd_data"].data.tobytes()) > 0
    assert str.encode("COP1") in raw["cmd_data"].data.tobytes()
    assert len(raw["weight_data"].data.tobytes()) > 0
    assert raw["weight_region"] == 0
    assert all(v > 0 for v in raw["scratch_shape"].tolist())
    assert raw["scratch_region"] == 1
    assert raw["scratch_size"] > 0
    assert raw["scratch_fast_shape"].tolist() == [0]
    assert raw["scratch_fast_region"] == 0
    assert raw["scratch_fast_size"] == 0
    assert raw["input_shape"][0].tolist() == [1, 1, 1, 64, 64, 16]
    assert raw["input_elem_size"][0] == 1
    assert raw["input_region"][0] == 1
    assert raw["input_offset"][0] >= 0
    assert raw["output_shape"][0].tolist() == [1, 1, 1, 64, 64, 8]
    assert raw["output_elem_size"][0] == 1
    assert raw["output_region"][0] == 1
    assert raw["output_offset"][0] >= 0
    # Model contains no variables, check that the values are empty arrays.
    assert raw["variable_shape"].size == 0
    assert raw["variable_elem_size"].size == 0
    assert raw["variable_region"].size == 0
    assert raw["variable_offset"].size == 0


def test_regor():
    """Testing regor module."""
    from ethosu import regor

    assert "Regor" in dir(regor)
    assert "compile" in dir(regor)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ClampAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ClampAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsClampAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ClampAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ClampAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ClampAttribute
    def MinInt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ClampAttribute
    def MaxInt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ClampAttribute
    def MinFp(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # ClampAttribute
    def MinFpAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # ClampAttribute
    def MinFpLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ClampAttribute
    def MinFpIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # ClampAttribute
    def MaxFp(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # ClampAttribute
    def MaxFpAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # ClampAttribute
    def MaxFpLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ClampAttribute
    def MaxFpIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

def ClampAttributeStart(builder):
    builder.StartObject(4)

def Start(builder):
    ClampAttributeStart(builder)

def ClampAttributeAddMinInt(builder, minInt):
    builder.PrependInt32Slot(0, minInt, 0)

def AddMinInt(builder, minInt):
    ClampAttributeAddMinInt(builder, minInt)

def ClampAttributeAddMaxInt(builder, maxInt):
    builder.PrependInt32Slot(1, maxInt, 0)

def AddMaxInt(builder, maxInt):
    ClampAttributeAddMaxInt(builder, maxInt)

def ClampAttributeAddMinFp(builder, minFp):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(minFp), 0)

def AddMinFp(builder, minFp):
    ClampAttributeAddMinFp(builder, minFp)

def ClampAttributeStartMinFpVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)

def StartMinFpVector(builder, numElems: int) -> int:
    return ClampAttributeStartMinFpVector(builder, numElems)

def ClampAttributeAddMaxFp(builder, maxFp):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(maxFp), 0)

def AddMaxFp(builder, maxFp):
    ClampAttributeAddMaxFp(builder, maxFp)

def ClampAttributeStartMaxFpVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)

def StartMaxFpVector(builder, numElems: int) -> int:
    return ClampAttributeStartMaxFpVector(builder, numElems)

def ClampAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ClampAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class CustomAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CustomAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCustomAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def CustomAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # CustomAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CustomAttribute
    def Identifier(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CustomAttribute
    def Config(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CustomAttribute
    def ImplementationAttrs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # CustomAttribute
    def ImplementationAttrsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # CustomAttribute
    def ImplementationAttrsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # CustomAttribute
    def ImplementationAttrsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def CustomAttributeStart(builder):
    builder.StartObject(3)

def Start(builder):
    CustomAttributeStart(builder)

def CustomAttributeAddIdentifier(builder, identifier):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(identifier), 0)

def AddIdentifier(builder, identifier):
    CustomAttributeAddIdentifier(builder, identifier)

def CustomAttributeAddConfig(builder, config):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(config), 0)

def AddConfig(builder, config):
    CustomAttributeAddConfig(builder, config)

def CustomAttributeAddImplementationAttrs(builder, implementationAttrs):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(implementationAttrs), 0)

def AddImplementationAttrs(builder, implementationAttrs):
    CustomAttributeAddImplementationAttrs(builder, implementationAttrs)

def CustomAttributeStartImplementationAttrsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)

def StartImplementationAttrsVector(builder, numElems: int) -> int:
    return CustomAttributeStartImplementationAttrsVector(builder, numElems)

def CustomAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return CustomAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TopKV2Options(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TopKV2Options()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTopKV2Options(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def TopKV2OptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # TopKV2Options
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

def TopKV2OptionsStart(builder):
    builder.StartObject(0)

def Start(builder):
    TopKV2OptionsStart(builder)

def TopKV2OptionsEnd(builder):
    return builder.EndObject()

def End(builder):
    return TopKV2OptionsEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class CondIfAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CondIfAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCondIfAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def CondIfAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # CondIfAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CondIfAttribute
    def ThenBranch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CondIfAttribute
    def ElseBranch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def CondIfAttributeStart(builder):
    builder.StartObject(2)

def Start(builder):
    CondIfAttributeStart(builder)

def CondIfAttributeAddThenBranch(builder, thenBranch):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(thenBranch), 0)

def AddThenBranch(builder, thenBranch):
    CondIfAttributeAddThenBranch(builder, thenBranch)

def CondIfAttributeAddElseBranch(builder, elseBranch):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(elseBranch), 0)

def AddElseBranch(builder, elseBranch):
    CondIfAttributeAddElseBranch(builder, elseBranch)

def CondIfAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return CondIfAttributeEnd(builder)



# SPDX-FileCopyrightText: Copyright 2020-2022, 2024-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
# SPDX-FileCopyrightText: (c) Meta Platforms, Inc. and affiliates. (http://www.meta.com)
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the License); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Description:
# Writes out per-pass and summary performance statistics to CSV files.
import csv
import sys

import numpy as np

from .nn_graph import PassPlacement
from .npu_performance import BandwidthDirection
from .npu_performance import PassCycles
from .numeric_util import round_up_to_int
from .operation import Op
from .tensor import MemArea
from .tensor import TensorPurpose


def mem_areas_to_report():
    # Exclude SHRAM, as the SHRAM performance numbers only cover LUT usage
    return [area for area in MemArea.all() if area != MemArea.Shram]


def write_summary_metrics_csv_common(
    summary_filename,
    arch,
    name,
    total_original_weights,
    total_npu_encoded_weights,
    n_passes,
    n_cascaded_passes,
    cycles,
    bandwidths,
    memory_used,
    batch_size,
    macs,
):
    with open(summary_filename, "w") as f:
        writer = csv.writer(f)

        mem_areas = mem_areas_to_report()

        labels = [
            "experiment",
            "network",
        ]

        labels += (
            ["accelerator_configuration", "system_config", "memory_mode", "core_clock", "arena_cache_size"]
            + [area.identifier_name() + "_bandwidth" for area in mem_areas]
            + ["weights_storage_area", "feature_map_storage_area"]
        )

        labels += [
            "inferences_per_second",
            "batch_size",
            "inference_time",
            "passes_before_fusing",
            "passes_after_fusing",
        ]
        labels += [area.identifier_name() + "_memory_used" for area in mem_areas]
        labels += ["total_original_weights"]
        labels += ["total_npu_encoded_weights"]

        for mem_area in mem_areas:
            labels += [
                mem_area.identifier_name() + "_feature_map_read_bytes",
                mem_area.identifier_name() + "_feature_map_write_bytes",
                mem_area.identifier_name() + "_weight_read_bytes",
                mem_area.identifier_name() + "_weight_write_bytes",
                mem_area.identifier_name() + "_total_bytes",
            ]

        labels += ["nn_macs", "nn_tops"]

        labels += ["cycles_" + kind.identifier_name() for kind in PassCycles.all()]

        writer.writerow(labels)

        data_items = [
            "default",
            name,
        ]

        if arch:
            data_items += (
                [
                    arch.accelerator_config.name,
                    arch.system_config,
                    arch.memory_mode,
                    arch.core_clock,
                    arch.arena_cache_size / 1024,
                ]
                + [arch.memory_bandwidths_per_second[mem_area] / 1024.0 / 1024 / 1024 for mem_area in mem_areas]
                + [
                    arch.tensor_storage_mem_area[TensorPurpose.Weights].display_name(),
                    arch.tensor_storage_mem_area[TensorPurpose.FeatureMap].display_name(),
                ]
            )

        midpoint_inference_time = cycles[PassCycles.Total] / arch.core_clock
        if midpoint_inference_time > 0:
            midpoint_fps = 1 / midpoint_inference_time
        else:
            midpoint_fps = np.nan

        data_items += [midpoint_fps, batch_size, midpoint_inference_time, n_passes, n_cascaded_passes]
        data_items += [memory_used.get(mem_area, 0) / 1024.0 for mem_area in mem_areas]
        data_items += [total_original_weights]
        data_items += [total_npu_encoded_weights]

        for mem_area in mem_areas:
            bws = bandwidths[mem_area]
            total_bw = np.sum(bws)
            weight_bws = bws[TensorPurpose.Weights]
            fm_bws = bws[TensorPurpose.FeatureMap]
            data_items += [
                fm_bws[BandwidthDirection.Read],
                fm_bws[BandwidthDirection.Write],
                weight_bws[BandwidthDirection.Read],
                weight_bws[BandwidthDirection.Write],
                total_bw,
            ]

        data_items += [
            macs,
            macs * 2 * midpoint_fps / 1e12,
        ]

        data_items += [cycles[kind] for kind in PassCycles.all()]

        writer.writerow(data_items)


def write_summary_metrics_csv(nng, summary_filename, arch):
    n_passes = sum(len(sg.passes) for sg in nng.subgraphs)
    n_cascaded_passes = sum(len(sg.cascaded_passes) for sg in nng.subgraphs)
    write_summary_metrics_csv_common(
        summary_filename,
        arch,
        nng.name,
        nng.total_original_weights,
        nng.total_npu_encoded_weights,
        n_passes,
        n_cascaded_passes,
        nng.cycles,
        nng.bandwidths,
        nng.memory_used,
        nng.batch_size,
        nng.macs,
    )


def write_pass_metrics_csv(nng, pass_filename):

    with open(pass_filename, "w") as f:
        writer = csv.writer(f)

        purpose_list = (
            ("total", (TensorPurpose.Weights, TensorPurpose.FeatureMap)),
            ("weights", (TensorPurpose.Weights,)),
            ("feature_map", (TensorPurpose.FeatureMap,)),
        )

        direction_list = (
            ("total", (BandwidthDirection.Read, BandwidthDirection.Write)),
            ("read", (BandwidthDirection.Read,)),
            ("write", (BandwidthDirection.Write,)),
        )
        bandwidth_names = []
        bandwidth_indices = []
        for mem_area in mem_areas_to_report():
            for purpose, purpose_candidates in purpose_list:
                for direction, direction_candidates in direction_list:
                    label = "bytes_{}_{}_{}".format(mem_area.identifier_name(), purpose, direction)
                    bandwidth_names.append(label)
                    bandwidth_indices.append((mem_area, purpose_candidates, direction_candidates))

        all_cycles = (
            PassCycles.Total,
            PassCycles.Npu,
            PassCycles.SramAccess,
            PassCycles.DramAccess,
            PassCycles.OnChipFlashAccess,
            PassCycles.OffChipFlashAccess,
        )
        writer.writerow(
            [
                "name",
                "operators",
                "placement",
                "streaming_strategy",
                "block_config_height",
                "block_config_width",
                "block_config_input_channels",
                "block_config_output_channels",
            ]
            + ["cycles_" + v.identifier_name() for v in all_cycles]
            + ["nn_macs"]
            + bandwidth_names
            + ["sram_used"]
        )

        def write_subgraph(sg):
            for cps in sg.cascaded_passes:
                if cps.placement == PassPlacement.StartupInit:
                    continue  # skip the dummy init pass

                for ps in cps.passes:
                    if len(ps.ops) == 1 and ps.ops[0].type == Op.CustomNpuOp:
                        # just treat this as a call, unroll it
                        write_subgraph(ps.ops[0].attrs["subgraph"])
                        continue
                    stats = [ps.name, " ".join(op.type.name for op in ps.ops)]
                    stats += [ps.placement.name]
                    stats += [cps.strategy.name]
                    stats += list(ps.block_config)
                    stats += [round_up_to_int(ps.cycles[v]) for v in all_cycles]
                    stats += [round_up_to_int(ps.macs)]
                    for indices in bandwidth_indices:
                        res = 0
                        i = indices[0]
                        for j in indices[1]:
                            for k in indices[2]:
                                res += round_up_to_int(ps.bandwidths[i, j, k])
                        stats.append(res)
                    try:
                        stats += [ps.sram_used]
                    except AttributeError:
                        stats += [0]

                    writer.writerow(stats)

        write_subgraph(nng.get_root_subgraph())


def print_performance_metrics_common(
    arch,
    name,
    cycles,
    macs,
    bandwidths,
    batch_size,
    memory_used,
    cpu_operations=None,
    npu_operations=None,
    show_cpu_operations=False,
    weights_data=None,
    verbose_cycle_estimate=False,
    f=sys.stdout,
):

    orig_mem_areas_labels = [(v, v.display_name()) for v in mem_areas_to_report()]

    inference_time = cycles[-1] / arch.core_clock
    if inference_time > 0:
        inferences_per_second = 1 / inference_time
    else:
        inferences_per_second = np.nan

    mem_area_labels = [
        (mem_area, label) for mem_area, label in orig_mem_areas_labels if np.sum(bandwidths[mem_area]) > 0
    ]

    print("", file=f)
    if name:
        print(f"Network summary for {name}", file=f)
    print(f"Accelerator configuration        {arch.accelerator_config.name:>20}", file=f)
    print(f"System configuration             {arch.system_config:>20}", file=f)
    print(f"Memory mode                      {arch.memory_mode:>20}", file=f)
    print(f"Accelerator clock                        {int(arch.core_clock / 1e6):12d} MHz", file=f)
    for mem_area, label in mem_area_labels:
        label += " bandwidth"
        bandwidth = arch.memory_bandwidths_per_second[mem_area] / 1024.0 / 1024 / 1024
        print(
            f"Design peak {label:25}    {bandwidth:12.2f} GB/s",
            file=f,
        )
    print(file=f)
    for mem_area, label in mem_area_labels:
        if mem_area not in memory_used:
            continue

        aug_label = label + " used"

        print(f"Total {aug_label:25}          {memory_used[mem_area] / 1024.0:12.2f} KiB", file=f)

    print(file=f)

    if cpu_operations is None:
        cpu_operations = []
    if npu_operations is None:
        npu_operations = []

    n_cpu_operations = len(cpu_operations)
    n_npu_operations = len(npu_operations)
    n_total_operations = max(n_cpu_operations + n_npu_operations, 1)  # avoid potential divide by zero

    for str_ops_type, n_ops, ops in (
        ("CPU", n_cpu_operations, cpu_operations),
        ("NPU", n_npu_operations, npu_operations),
    ):
        print(f"{str_ops_type} operators = {n_ops:d} ({n_ops / n_total_operations:4.1%})", file=f)
        if show_cpu_operations:
            for op in ops:
                print(f"   {str_ops_type}: {op}")

    print("", file=f)

    for mem_area, label in mem_area_labels:
        bws = bandwidths[mem_area]
        total_bw = np.sum(bws)
        weight_bws = bws[TensorPurpose.Weights]
        fm_bws = bws[TensorPurpose.FeatureMap]
        aug_label = label + " bandwidth"
        print(
            f"Average {aug_label:25}        {total_bw * inferences_per_second / 1024.0 / 1024.0 / 1000.0:12.2f} GB/s",
            file=f,
        )
        print(
            f"Input   {aug_label:25}        "
            f"{np.sum(fm_bws[BandwidthDirection.Read]) / 1024.0 / 1024.0:12.2f} MB/batch",
            file=f,
        )
        print(f"Weight  {aug_label:25}        {np.sum(weight_bws) / 1024.0 / 1024.0:12.2f} MB/batch", file=f)
        print(
            f"Output  {aug_label:25}        "
            f"{np.sum(fm_bws[BandwidthDirection.Write]) / 1024.0 / 1024.0:12.2f} MB/batch",
            file=f,
        )
        print(f"Total   {aug_label:25}        {total_bw / 1024.0 / 1024.0:12.2f} MB/batch", file=f)
        print(
            f"Total   {aug_label:25} per input "
            f"{total_bw / 1024.0 / 1024.0 / batch_size:9.2f} MB/inference (batch size {batch_size:d})",
            file=f,
        )
        print(file=f)

    if weights_data:
        print(f"Original Weights Size                    {weights_data['original'] / 1024.0:12.2f} KiB", file=f)
        print(f"NPU Encoded Weights Size                 {weights_data['npu_encoded'] / 1024.0:12.2f} KiB", file=f)
        print(file=f)

    print(
        f"Neural network macs                      {int(macs):12d} MACs/batch",
        file=f,
    )
    if verbose_cycle_estimate:
        print(file=f)
        print("Info: The numbers below are internal compiler estimates.", file=f)
        print("For performance numbers the compiled network should be run on an FVP Model or FPGA.", file=f)
        print(file=f)

        print(
            f"Network Tops/s                           {macs * 2 * inferences_per_second / 1e12:12.2f} Tops/s",
            file=f,
        )
        print(file=f)

        for kind in PassCycles.all():
            aug_label = kind.display_name() + " cycles"
            cyc = cycles[kind]
            print(f"{aug_label:30}           {int(cyc):12d} cycles/batch", file=f)
        print(file=f)

        print(
            f"Batch Inference time              {inference_time * 1000:7.2f} ms,"
            f" {inferences_per_second:7.2f} inferences/s (batch size {batch_size:d})",
            file=f,
        )
    print(file=f)


def print_performance_metrics(
    nng, arch, show_cpu_operations=False, verbose_weights=False, verbose_cycle_estimate=False, f=sys.stdout
):
    cpu_operations = []
    npu_operations = []
    ir_only_ops = (
        Op.Const,
        Op.Placeholder,
        Op.CustomNpuOp,
        Op.SubgraphInput,
    )

    def format_tens_list(lst):
        return " ".join(str(list(tens.shape)) for tens in lst if tens is not None)

    for sg in nng.subgraphs:
        if sg.placement == PassPlacement.Cpu:
            for op in sg.get_all_ops_from_passes():
                if op.type not in ir_only_ops:
                    cpu_operations.append(
                        f"{op.type} = {op.name} "
                        f"(inputs {format_tens_list(op.inputs)}, outputs {format_tens_list(op.outputs)})"
                    )
        elif sg.placement == PassPlacement.Npu:
            for op in sg.get_all_ops_from_passes():
                if op.type not in ir_only_ops:
                    npu_operations.append(
                        f"{op.type} = {op.name} "
                        f"(inputs {format_tens_list(op.inputs)}, outputs {format_tens_list(op.outputs)})"
                    )

    weights_data = (
        {"original": nng.total_original_weights, "npu_encoded": nng.total_npu_encoded_weights}
        if verbose_weights
        else None
    )

    return print_performance_metrics_common(
        arch,
        nng.name,
        nng.cycles,
        nng.macs,
        nng.bandwidths,
        nng.batch_size,
        nng.memory_used,
        cpu_operations,
        npu_operations,
        show_cpu_operations,
        weights_data,
        verbose_cycle_estimate,
        f,
    )


def regor_operations_from_database(opt_database):
    def find_in_header(names, header):
        idx = []
        for name in names:
            if name not in header:
                idx.append(-1)
            idx.append(header.index(name))
        return idx

    # set of optimised_ids that ended up on NPU
    npu_optimised_ids = set()
    # maps src-id to ofm-shape
    ofm_shapes = dict()
    cpu_operations = []
    npu_operations = []

    # prerequisite checks
    required_tables = ["source", "optimised", "perf", "queue"]
    for table in required_tables:
        if table not in opt_database.tables:
            print("Could not extract CPU operations:")
            print("Table: {} was not in opt_database".format(table))
            return cpu_operations, npu_operations

    # build ofm_shapes from source table
    src = opt_database.tables["source"]
    fields = ["id", "ofm_w", "ofm_h", "ofm_d"]
    ids = find_in_header(fields, src.header)
    if any([x < 0 for x in ids]):
        print("Could not extract CPU operations:")
        print("Could not find all necessary fields in source database")
        return cpu_operations, npu_operations

    id_idx, w_idx, h_idx, d_idx = ids
    for entry in src.data:
        src_id = entry[id_idx]
        w = int(entry[w_idx])
        h = int(entry[h_idx])
        d = int(entry[d_idx])
        ofm_shapes[src_id] = [1, w, h, d]

    # build npu_optimised_ids from queue table
    qt = opt_database.tables["queue"]
    id_idx = find_in_header(["optimised_id"], qt.header)[0]
    if id_idx == -1:
        print("Could not extract CPU operations:")
        print("optimised_id was not found in queue table")
        return cpu_operations, npu_operations
    for entry in qt.data:
        npu_optimised_ids.add(entry[id_idx])

    # build cpu/npu operations from perf-table
    perf = opt_database.tables["perf"]
    fields = ["optimised_id", "source_id", "name", "operator"]
    ids = find_in_header(fields, perf.header)
    if any([x < 0 for x in ids]):
        print("Could not extract CPU operations:")
        print("Could not find all necessary fields in perf database")
    opt_idx, src_idx, name_idx, operator_idx = ids
    for entry in perf.data:
        opt_id = entry[opt_idx]
        src_id = entry[src_idx]
        name = entry[name_idx]
        operator = entry[operator_idx]
        ofm_shape = ofm_shapes[src_id]
        # TODO add ifm shapes
        op_desc = f"{operator} = {name} (outputs {ofm_shape})"
        if opt_id in npu_optimised_ids:
            npu_operations.append(op_desc)
        else:
            cpu_operations.append(op_desc)
    return cpu_operations, npu_operations


def print_regor_performance_metrics(
    arch,
    report,
    model_name,
    csv_filename,
    opt_database,
    verbose_weights=False,
    verbose_cycle_estimate=False,
    show_cpu_operations=False,
):
    # Map from Regor memory names to Vela MemArea
    memory_mapping = {
        "sram": MemArea.Sram,
        "dram": MemArea.Dram,
        "onchipflash": MemArea.OnChipFlash,
        "offchipflash": MemArea.OffChipFlash,
        "lutram": MemArea.Shram,
        "shram": MemArea.Shram,
    }

    # Map from Regor memory names to Vela PassCycles
    cycles_mapping = {
        "sram": PassCycles.SramAccess,
        "dram": PassCycles.DramAccess,
        "onchipflash": PassCycles.OnChipFlashAccess,
        "offchipflash": PassCycles.OffChipFlashAccess,
    }

    # Map from Regor access_type to Vela TensorPurpose
    purpose_mapping = {
        "lut": TensorPurpose.LUT,
        "featuremap": TensorPurpose.FeatureMap,
        "scale": TensorPurpose.FSBias,
        "weights": TensorPurpose.Weights,
    }

    cycles = [0] * PassCycles.Size
    bandwidths = [[[0] * BandwidthDirection.Size for i in range(TensorPurpose.Size)] for j in range(MemArea.Size)]
    memory_used = {i: 0 for i in range(MemArea.Size)}
    for mem_name, memory in report.memories.items():
        for _, a in memory.accesses.items():
            mem_name_lower = str(mem_name).lower()

            # skip shram/lutram in performance report
            if mem_name_lower in ["lutram", "shram"]:
                continue

            mem_area = memory_mapping.get(mem_name_lower, MemArea.Unknown)
            purpose = purpose_mapping.get(str(a.accessType).lower(), TensorPurpose.Unknown)

            if mem_name_lower in cycles_mapping:
                cycles_idx = cycles_mapping[mem_name_lower]
                cycles[cycles_idx] += a.accessCycles

            bandwidths[mem_area][purpose][BandwidthDirection.Read] = a.bytesRead
            bandwidths[mem_area][purpose][BandwidthDirection.Write] = a.bytesWritten

            memory_used[mem_area] = memory.peakUsage

    cycles[PassCycles.Npu] = report.npuCycles
    cycles[PassCycles.Total] = report.totalCycles

    batch_size = 1

    npu_operations = list()
    cpu_operations = list()

    if opt_database:
        if show_cpu_operations:
            cpu_operations, npu_operations = regor_operations_from_database(opt_database)

    if not len(cpu_operations) and not len(npu_operations):
        cpu_operations = ["N/A"] * report.cpuOps
        npu_operations = ["N/A"] * report.npuOps

    total_original_weights = report.originalWeights
    total_npu_encoded_weights = report.encodedWeights

    weights_data = (
        {"original": total_original_weights, "npu_encoded": total_npu_encoded_weights} if verbose_weights else None
    )
    f = sys.stdout

    print_performance_metrics_common(
        arch,
        model_name,
        cycles,
        report.macCount,
        bandwidths,
        batch_size,
        memory_used,
        cpu_operations,
        npu_operations,
        show_cpu_operations,
        weights_data,
        verbose_cycle_estimate,
        f,
    )

    n_passes = report.cpuOps + report.npuOps
    n_passes_after_cascading = n_passes - report.cascadedOps + report.cascades

    write_summary_metrics_csv_common(
        csv_filename,
        arch,
        model_name,
        total_original_weights,
        total_npu_encoded_weights,
        n_passes,
        n_passes_after_cascading,
        cycles,
        bandwidths,
        memory_used,
        batch_size,
        report.macCount,
    )


def postprocess_regor_performance_database(arch, opt_database, performance_report):
    if "perf" not in opt_database.tables:
        return None, None

    def find_in_header(name, header):
        if name not in header:
            return -1
        return header.index(name)

    def _percentage(x, y):
        if y == 0:
            return 100
        return (x * 100) / y

    # Set of operations in the optimised-table that ended up as commands
    npu_operators = set()
    if "queue" in opt_database.tables:
        queue_table = opt_database.tables["queue"]
        queue_header = queue_table.header
        queue_rows = queue_table.data
        opt_idx = find_in_header("optimised_id", queue_header)
        if opt_idx != -1:
            for row in queue_rows:
                npu_operators.add(int(row[opt_idx]))

    perf_table = opt_database.tables["perf"]
    perf_header = perf_table.header
    perf_rows = perf_table.data

    # populate id -> name map for source operations
    source_names = {}

    if "source" in opt_database.tables:
        source_table = opt_database.tables["source"]
        source_header = source_table.header
        source_rows = source_table.data
        op_name_idx = find_in_header("operator", source_header)
        op_id_idx = find_in_header("id", source_header)
        if op_name_idx != -1 and op_id_idx != -1:
            for row in source_rows:
                op_id = row[op_id_idx]
                op_name = row[op_name_idx]
                source_names[op_id] = op_name

    # maps vela-columns to columns in the regor performance-database
    col_map = {
        "SourceId": find_in_header("source_id", perf_header),
        "OptId": find_in_header("optimised_id", perf_header),
        "NNG Operator": find_in_header("operator", perf_header),
        "Staging Usage": find_in_header("staging_usage", perf_header),
        "Op Cycles": find_in_header("op_cycles", perf_header),
        "NPU": find_in_header("npu_cycles", perf_header),
        "MAC Count": find_in_header("mac_count", perf_header),
        "SRAM AC": find_in_header("Sram_ac", perf_header),
        "DRAM AC": find_in_header("Dram_ac", perf_header),
        "OnFlash AC": find_in_header("OnChipFlash_ac", perf_header),
        "OffFlash AC": find_in_header("OffChipFlash_ac", perf_header),
        "Name": find_in_header("name", perf_header),
    }

    # generate new header and data that aligns with vela format
    new_header = [
        "Original Operator",
        "NNG Operator",
        "Target",
        "Staging Usage",
        "Peak% (Staging)",
        "Op Cycles",
        "Network% (cycles)",
        "NPU",
        "SRAM AC",
        "DRAM AC",
        "OnFlash AC",
        "OffFlash AC",
        "MAC Count",
        "Network% (MAC)",
        "Util% (MAC)",
        "Name",
    ]
    new_data = []
    for row in perf_rows:
        new_row = []
        for col in new_header:
            if col in col_map:
                idx = col_map[col]
                if idx == -1:
                    new_row.append("0")
                else:
                    new_row.append(row[idx])
            # post-processing for things not contained in regors performance-database
            else:
                if col == "Target":
                    nng_op_idx = col_map["NNG Operator"]
                    if nng_op_idx != -1:
                        nng_op = row[nng_op_idx]
                        if nng_op != "Passthrough":
                            new_row.append("NPU")
                        else:
                            new_row.append("CPU")
                    else:
                        new_row.append("CPU")
                elif col == "Original Operator":
                    source_id_idx = col_map["SourceId"]
                    if source_id_idx != -1:
                        source_op_id = row[source_id_idx]
                        source_name = source_names.get(source_op_id, "N/A")
                        new_row.append(source_name)
                    else:
                        new_row.append("N/A")
                elif col == "Peak% (Staging)":
                    staging_idx = col_map["Staging Usage"]
                    if staging_idx != -1 and performance_report.stagingMemoryArea in performance_report.memories:
                        staging_usage = int(row[staging_idx])
                        staging_network = int(
                            performance_report.memories[performance_report.stagingMemoryArea].peakUsage
                        )
                        staging_percent = _percentage(staging_usage, staging_network)
                        new_row.append(str(staging_percent))
                    else:
                        new_row.append("NaN")
                elif col == "Network% (cycles)":
                    cycles_idx = col_map["Op Cycles"]
                    if cycles_idx != -1:
                        cycles = int(row[cycles_idx])
                        cycles_network = int(performance_report.totalCycles)
                        cycles_percent = _percentage(cycles, cycles_network)
                        new_row.append(str(cycles_percent))
                    else:
                        new_row.append("NaN")
                elif col == "Network% (MAC)":
                    mac_idx = col_map["MAC Count"]
                    if mac_idx != -1:
                        macs = int(row[mac_idx])
                        macs_network = int(performance_report.macCount)
                        macs_percent = _percentage(macs, macs_network)
                        new_row.append(str(macs_percent))
                    else:
                        new_row.append("100")
                elif col == "Util% (MAC)":
                    cycles_idx = col_map["Op Cycles"]
                    mac_idx = col_map["MAC Count"]
                    if cycles_idx != -1 and mac_idx != -1:
                        cycles = int(row[cycles_idx])
                        macs = int(row[mac_idx])
                        max_macs = cycles * arch.num_macs_per_cycle * arch.ncores
                        util = _percentage(macs, max_macs)
                        new_row.append(str(util))
                    else:
                        new_row.append("NaN")
                else:
                    new_row.append("NaN")

        new_data.append(new_row)
    return new_header, new_data


def write_regor_db(opt_database, output_basename):
    out = "<?xml version='1.0' encoding='UTF-8'?>\n"
    out += f'<debug source="{output_basename}">\n'
    for name, table in opt_database.tables.items():
        out += f'<table name="{name}">\n'
        out += "<![CDATA[\n"
        out += ",".join(f'"{col}"' for col in table.header) + "\n"
        for row in table.data:
            for i in range(len(row)):
                if not row[i].isnumeric():
                    row[i] = f'"{row[i]}"'
            out += ",".join(f"{val}" for val in row) + "\n"
        out += "]]>\n"
        out += "</table>\n"
    out += "</debug>\n"

    debug_filename = output_basename + "_debug.xml"

    with open(debug_filename, "w", encoding="utf-8") as file:
        file.write(out)


def write_regor_perlayer_performance_csv(arch, opt_database, performance_report, output_basename):
    if "perf" not in opt_database.tables:
        return

    perf_header, perf_rows = postprocess_regor_performance_database(arch, opt_database, performance_report)

    csv_name = f"{output_basename}_per-layer.csv"
    with open(csv_name, "w") as csv_file:
        csv_file.write(",".join(perf_header))
        csv_file.write("\n")
        for row in perf_rows:
            csv_file.write(",".join(row))
            csv_file.write("\n")


def print_regor_perlayer_performance(arch, opt_database, performance_report, output_basename):
    if "perf" not in opt_database.tables:
        return

    perf_header, perf_rows = postprocess_regor_performance_database(arch, opt_database, performance_report)

    # header -> (align, width, precision)
    header = {
        "Original Operator": ("<", 20, -1),
        "NNG Operator": ("<", 20, -1),
        "Target": ("<", 6, -1),
        "Staging Usage": (">", 13, 0.0),
        "Peak% (Staging)": (">", 16, 0.2),
        "Op Cycles": (">", 10, 0.0),
        "Network% (cycles)": (">", 17, 0.2),
        "NPU": (">", 10, 0.0),
        "SRAM AC": (">", 10, 0.0),
        "DRAM AC": (">", 10, 0.0),
        "OnFlash AC": (">", 10, 0.0),
        "OffFlash AC": (">", 11, 0.0),
        "MAC Count": (">", 10, 0.0),
        "Network% (MAC)": (">", 14, 0.2),
        "Util% (MAC)": (">", 12, 0.2),
        "Name": ("<", 20, -1),
    }

    print(f"\n{str('#') * 80}\nPerformance for NPU Grap {output_basename}")

    line = ""
    line2 = ""
    for col_name in header:
        align, width, _ = header[col_name]
        line_data = f"{col_name:{align}{width}}"
        line += line_data + " "
        line2 += "-" * len(line_data) + " "
    print(line)
    print(line2)

    for op_data in perf_rows:
        line = ""
        for table_idx in range(len(perf_header)):
            h = perf_header[table_idx]
            align, width, precision = header[h]
            if precision == -1:
                w = str(width)
            else:
                w = str(width + precision) + "f"
            item = op_data[table_idx]
            if precision != -1:
                item = float(item)
            line += f"{item:{align}{w}}" + " "
        print(line)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ResizeAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ResizeAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsResizeAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ResizeAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ResizeAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ResizeAttribute
    def Scale(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int16Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 2))
        return 0

    # ResizeAttribute
    def ScaleAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int16Flags, o)
        return 0

    # ResizeAttribute
    def ScaleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ResizeAttribute
    def ScaleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # ResizeAttribute
    def Offset(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int16Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 2))
        return 0

    # ResizeAttribute
    def OffsetAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int16Flags, o)
        return 0

    # ResizeAttribute
    def OffsetLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ResizeAttribute
    def OffsetIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # ResizeAttribute
    def Border(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int16Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 2))
        return 0

    # ResizeAttribute
    def BorderAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int16Flags, o)
        return 0

    # ResizeAttribute
    def BorderLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ResizeAttribute
    def BorderIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # ResizeAttribute
    def Mode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def ResizeAttributeStart(builder):
    builder.StartObject(4)

def Start(builder):
    ResizeAttributeStart(builder)

def ResizeAttributeAddScale(builder, scale):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(scale), 0)

def AddScale(builder, scale):
    ResizeAttributeAddScale(builder, scale)

def ResizeAttributeStartScaleVector(builder, numElems):
    return builder.StartVector(2, numElems, 2)

def StartScaleVector(builder, numElems: int) -> int:
    return ResizeAttributeStartScaleVector(builder, numElems)

def ResizeAttributeAddOffset(builder, offset):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(offset), 0)

def AddOffset(builder, offset):
    ResizeAttributeAddOffset(builder, offset)

def ResizeAttributeStartOffsetVector(builder, numElems):
    return builder.StartVector(2, numElems, 2)

def StartOffsetVector(builder, numElems: int) -> int:
    return ResizeAttributeStartOffsetVector(builder, numElems)

def ResizeAttributeAddBorder(builder, border):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(border), 0)

def AddBorder(builder, border):
    ResizeAttributeAddBorder(builder, border)

def ResizeAttributeStartBorderVector(builder, numElems):
    return builder.StartVector(2, numElems, 2)

def StartBorderVector(builder, numElems: int) -> int:
    return ResizeAttributeStartBorderVector(builder, numElems)

def ResizeAttributeAddMode(builder, mode):
    builder.PrependUint32Slot(3, mode, 0)

def AddMode(builder, mode):
    ResizeAttributeAddMode(builder, mode)

def ResizeAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ResizeAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TosaBasicBlock(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TosaBasicBlock()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTosaBasicBlock(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def TosaBasicBlockBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # TosaBasicBlock
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TosaBasicBlock
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # TosaBasicBlock
    def Operators(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .TosaOperator import TosaOperator
            obj = TosaOperator()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TosaBasicBlock
    def OperatorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaBasicBlock
    def OperatorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # TosaBasicBlock
    def Tensors(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .TosaTensor import TosaTensor
            obj = TosaTensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TosaBasicBlock
    def TensorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaBasicBlock
    def TensorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # TosaBasicBlock
    def Inputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # TosaBasicBlock
    def InputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaBasicBlock
    def InputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # TosaBasicBlock
    def Outputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # TosaBasicBlock
    def OutputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaBasicBlock
    def OutputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def TosaBasicBlockStart(builder):
    builder.StartObject(5)

def Start(builder):
    TosaBasicBlockStart(builder)

def TosaBasicBlockAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    TosaBasicBlockAddName(builder, name)

def TosaBasicBlockAddOperators(builder, operators):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(operators), 0)

def AddOperators(builder, operators):
    TosaBasicBlockAddOperators(builder, operators)

def TosaBasicBlockStartOperatorsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOperatorsVector(builder, numElems: int) -> int:
    return TosaBasicBlockStartOperatorsVector(builder, numElems)

def TosaBasicBlockAddTensors(builder, tensors):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(tensors), 0)

def AddTensors(builder, tensors):
    TosaBasicBlockAddTensors(builder, tensors)

def TosaBasicBlockStartTensorsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartTensorsVector(builder, numElems: int) -> int:
    return TosaBasicBlockStartTensorsVector(builder, numElems)

def TosaBasicBlockAddInputs(builder, inputs):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(inputs), 0)

def AddInputs(builder, inputs):
    TosaBasicBlockAddInputs(builder, inputs)

def TosaBasicBlockStartInputsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInputsVector(builder, numElems: int) -> int:
    return TosaBasicBlockStartInputsVector(builder, numElems)

def TosaBasicBlockAddOutputs(builder, outputs):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(outputs), 0)

def AddOutputs(builder, outputs):
    TosaBasicBlockAddOutputs(builder, outputs)

def TosaBasicBlockStartOutputsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOutputsVector(builder, numElems: int) -> int:
    return TosaBasicBlockStartOutputsVector(builder, numElems)

def TosaBasicBlockEnd(builder):
    return builder.EndObject()

def End(builder):
    return TosaBasicBlockEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class MulAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MulAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMulAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def MulAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # MulAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MulAttribute
    def Shift(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def MulAttributeStart(builder):
    builder.StartObject(1)

def Start(builder):
    MulAttributeStart(builder)

def MulAttributeAddShift(builder, shift):
    builder.PrependInt32Slot(0, shift, 0)

def AddShift(builder, shift):
    MulAttributeAddShift(builder, shift)

def MulAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return MulAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TosaOperator(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TosaOperator()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTosaOperator(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def TosaOperatorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # TosaOperator
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TosaOperator
    def Op(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # TosaOperator
    def AttributeType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # TosaOperator
    def Attribute(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # TosaOperator
    def Inputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # TosaOperator
    def InputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaOperator
    def InputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # TosaOperator
    def Outputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # TosaOperator
    def OutputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TosaOperator
    def OutputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def TosaOperatorStart(builder):
    builder.StartObject(5)

def Start(builder):
    TosaOperatorStart(builder)

def TosaOperatorAddOp(builder, op):
    builder.PrependUint32Slot(0, op, 0)

def AddOp(builder, op):
    TosaOperatorAddOp(builder, op)

def TosaOperatorAddAttributeType(builder, attributeType):
    builder.PrependUint8Slot(1, attributeType, 0)

def AddAttributeType(builder, attributeType):
    TosaOperatorAddAttributeType(builder, attributeType)

def TosaOperatorAddAttribute(builder, attribute):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(attribute), 0)

def AddAttribute(builder, attribute):
    TosaOperatorAddAttribute(builder, attribute)

def TosaOperatorAddInputs(builder, inputs):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(inputs), 0)

def AddInputs(builder, inputs):
    TosaOperatorAddInputs(builder, inputs)

def TosaOperatorStartInputsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInputsVector(builder, numElems: int) -> int:
    return TosaOperatorStartInputsVector(builder, numElems)

def TosaOperatorAddOutputs(builder, outputs):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(outputs), 0)

def AddOutputs(builder, outputs):
    TosaOperatorAddOutputs(builder, outputs)

def TosaOperatorStartOutputsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOutputsVector(builder, numElems: int) -> int:
    return TosaOperatorStartOutputsVector(builder, numElems)

def TosaOperatorEnd(builder):
    return builder.EndObject()

def End(builder):
    return TosaOperatorEnd(builder)



#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: Copyright 2023 Arm Limited and/or its affiliates <open-source-office@arm.com>
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the License); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Cppcheck wrapper."""
import json
import os
import subprocess
import sys

# Load compile database
def get_db_cmd(src):
    """Load compile database, looks for src compile options and filters them."""
    fcmd = []
    compile_db = None
    if "CMAKE_BINARY_DIR" in os.environ:
        compile_db = os.path.join(os.environ["CMAKE_BINARY_DIR"], "compile_commands.json")
        if not os.path.exists(compile_db):
            compile_db = None
    if compile_db:
        with open(compile_db) as f:
            cmds = json.load(f)
        for cmd in cmds:
            if cmd["file"] == src:
                cmd = cmd["command"].split()
                i = 0
                while i < len(cmd):
                    if cmd[i].startswith("-std"):
                        fcmd.append("-" + cmd[i])
                    if cmd[i].startswith("-D"):
                        fcmd.append(cmd[i])
                    if cmd[i].startswith("-isystem"):
                        fcmd.append("-I")
                        i += 1
                        fcmd.append(cmd[i])
                        fcmd.append(f"--suppress=*:{cmd[i]}/*")
                        fcmd.append("--suppress=unmatchedSuppression:*")
                    if cmd[i].startswith("-I"):
                        fcmd.append(cmd[i])
                        i += 1
                        fcmd.append(cmd[i])
                    i += 1
                break
    return fcmd


# The source file to scan is the last argument
src = sys.argv[-1]
assert os.path.exists(src), f"File {src} not found"

cmd = []
# CMake options
cmd += sys.argv[1:-1]
# Options from compile database
cmd += get_db_cmd(src)
# The source file last
cmd.append(src)

subprocess.call(cmd)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ArithmeticRightShiftAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ArithmeticRightShiftAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsArithmeticRightShiftAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ArithmeticRightShiftAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ArithmeticRightShiftAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ArithmeticRightShiftAttribute
    def Round(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def ArithmeticRightShiftAttributeStart(builder):
    builder.StartObject(1)

def Start(builder):
    ArithmeticRightShiftAttributeStart(builder)

def ArithmeticRightShiftAttributeAddRound(builder, round):
    builder.PrependBoolSlot(0, round, 0)

def AddRound(builder, round):
    ArithmeticRightShiftAttributeAddRound(builder, round)

def ArithmeticRightShiftAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ArithmeticRightShiftAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class PadAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PadAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPadAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def PadAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # PadAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PadAttribute
    def Padding(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # PadAttribute
    def PaddingAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # PadAttribute
    def PaddingLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # PadAttribute
    def PaddingIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # PadAttribute
    def PadConstInt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PadAttribute
    def PadConstFp(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # PadAttribute
    def PadConstFpAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # PadAttribute
    def PadConstFpLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # PadAttribute
    def PadConstFpIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def PadAttributeStart(builder):
    builder.StartObject(3)

def Start(builder):
    PadAttributeStart(builder)

def PadAttributeAddPadding(builder, padding):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(padding), 0)

def AddPadding(builder, padding):
    PadAttributeAddPadding(builder, padding)

def PadAttributeStartPaddingVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartPaddingVector(builder, numElems: int) -> int:
    return PadAttributeStartPaddingVector(builder, numElems)

def PadAttributeAddPadConstInt(builder, padConstInt):
    builder.PrependInt32Slot(1, padConstInt, 0)

def AddPadConstInt(builder, padConstInt):
    PadAttributeAddPadConstInt(builder, padConstInt)

def PadAttributeAddPadConstFp(builder, padConstFp):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(padConstFp), 0)

def AddPadConstFp(builder, padConstFp):
    PadAttributeAddPadConstFp(builder, padConstFp)

def PadAttributeStartPadConstFpVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)

def StartPadConstFpVector(builder, numElems: int) -> int:
    return PadAttributeStartPadConstFpVector(builder, numElems)

def PadAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return PadAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SliceAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SliceAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSliceAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def SliceAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # SliceAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SliceAttribute
    def Start(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SliceAttribute
    def StartAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SliceAttribute
    def StartLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SliceAttribute
    def StartIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # SliceAttribute
    def Size(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SliceAttribute
    def SizeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SliceAttribute
    def SizeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SliceAttribute
    def SizeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def SliceAttributeStart(builder):
    builder.StartObject(2)

def Start(builder):
    SliceAttributeStart(builder)

def SliceAttributeAddStart(builder, start):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(start), 0)

def AddStart(builder, start):
    SliceAttributeAddStart(builder, start)

def SliceAttributeStartStartVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartStartVector(builder, numElems: int) -> int:
    return SliceAttributeStartStartVector(builder, numElems)

def SliceAttributeAddSize(builder, size):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(size), 0)

def AddSize(builder, size):
    SliceAttributeAddSize(builder, size)

def SliceAttributeStartSizeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartSizeVector(builder, numElems: int) -> int:
    return SliceAttributeStartSizeVector(builder, numElems)

def SliceAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return SliceAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ReshapeAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ReshapeAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsReshapeAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ReshapeAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ReshapeAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ReshapeAttribute
    def NewShape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ReshapeAttribute
    def NewShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ReshapeAttribute
    def NewShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ReshapeAttribute
    def NewShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def ReshapeAttributeStart(builder):
    builder.StartObject(1)

def Start(builder):
    ReshapeAttributeStart(builder)

def ReshapeAttributeAddNewShape(builder, newShape):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(newShape), 0)

def AddNewShape(builder, newShape):
    ReshapeAttributeAddNewShape(builder, newShape)

def ReshapeAttributeStartNewShapeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartNewShapeVector(builder, numElems: int) -> int:
    return ReshapeAttributeStartNewShapeVector(builder, numElems)

def ReshapeAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ReshapeAttributeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ConvAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ConvAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsConvAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ConvAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ConvAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ConvAttribute
    def Pad(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ConvAttribute
    def PadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ConvAttribute
    def PadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ConvAttribute
    def PadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # ConvAttribute
    def Stride(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ConvAttribute
    def StrideAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ConvAttribute
    def StrideLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ConvAttribute
    def StrideIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # ConvAttribute
    def Dilation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ConvAttribute
    def DilationAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ConvAttribute
    def DilationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ConvAttribute
    def DilationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # ConvAttribute
    def InputZp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ConvAttribute
    def WeightZp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def ConvAttributeStart(builder):
    builder.StartObject(5)

def Start(builder):
    ConvAttributeStart(builder)

def ConvAttributeAddPad(builder, pad):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(pad), 0)

def AddPad(builder, pad):
    ConvAttributeAddPad(builder, pad)

def ConvAttributeStartPadVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartPadVector(builder, numElems: int) -> int:
    return ConvAttributeStartPadVector(builder, numElems)

def ConvAttributeAddStride(builder, stride):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(stride), 0)

def AddStride(builder, stride):
    ConvAttributeAddStride(builder, stride)

def ConvAttributeStartStrideVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartStrideVector(builder, numElems: int) -> int:
    return ConvAttributeStartStrideVector(builder, numElems)

def ConvAttributeAddDilation(builder, dilation):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(dilation), 0)

def AddDilation(builder, dilation):
    ConvAttributeAddDilation(builder, dilation)

def ConvAttributeStartDilationVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartDilationVector(builder, numElems: int) -> int:
    return ConvAttributeStartDilationVector(builder, numElems)

def ConvAttributeAddInputZp(builder, inputZp):
    builder.PrependInt32Slot(3, inputZp, 0)

def AddInputZp(builder, inputZp):
    ConvAttributeAddInputZp(builder, inputZp)

def ConvAttributeAddWeightZp(builder, weightZp):
    builder.PrependInt32Slot(4, weightZp, 0)

def AddWeightZp(builder, weightZp):
    ConvAttributeAddWeightZp(builder, weightZp)

def ConvAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ConvAttributeEnd(builder)
